package com.molino.mills.agent;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import com.molino.mills.model.BitBoardUtil;
import com.molino.mills.model.Board;
import com.molino.mills.model.BoardFeatures;
import com.molino.mills.model.GameState;
import com.molino.mills.model.MoveGenerator;

public class BoardAnalyser
{
    static final Logger log = Logger.getLogger(BoardAnalyser.class);

    static final int[] THREE_WAY_INTERSECTIONS = { 1, 3, 5, 7, 17, 19, 21, 23 };
    static final int THREE_WAY_INTERSECTIONS_BB = BitBoardUtil.positionsToBitBoard(THREE_WAY_INTERSECTIONS);
    static final int[] FOUR_WAY_INTERSECTIONS = { 9, 11, 13, 15 };
    static final int FOUR_WAY_INTERSECTIONS_BB = BitBoardUtil.positionsToBitBoard(FOUR_WAY_INTERSECTIONS);

    static final int WIN_VALUE = 1000000;
    static final int FOUR_WAY_INTERSECTION_VALUE = 64;
    static final int THREE_WAY_INTERSECTION_VALUE = 16;
    static final int SPOKE_EXCLUSIVITY_VALUE = 24;
    static final int MIDDLE_RING_EXCLUSIVITY_VALUE = 16;
    static final int MILL_VALUE = 32;

    static final int ROW_EXCLUSIVITY_VALUE = 64;

    static final int POTENTIAL_MILL_VALUE = 160;

    static final int POTENTIAL_MILL_NEXT_TURN_VALUE = 128;


    static final int PIECE_ADVANTAGE_EXPONENT = 8;
    static final int MOBILITY_ADVANTAGE_EXPONENT = 6;
    static final int MILL_MOBILITY_EXPONENT = 3;
    static final int SINGLE_JUMPER_EXPONENT = 3;

    static Level originalLogLevel = log.getLevel();

    public static int calculateScore(GameState state)
    {
        int player = state.currentPlayer;
        int score;
        Board board = state.board;
        int otherPlayer = 1 - player;
        int freebb = state.features.freebb;

  
        if (state.board.piecesLeft[player] < 3)
            return -BoardAnalyser.WIN_VALUE;
        else if (state.board.piecesLeft[otherPlayer] < 3) 
            return BoardAnalyser.WIN_VALUE;      
        

        score = BoardAnalyser.calculatePieceAdvantage(state.board, player)  << BoardAnalyser.PIECE_ADVANTAGE_EXPONENT;
        
        int jumping = (state.board.piecesLeft[0] == 3 ? 1 : 0) + 
                      (state.board.piecesLeft[1] == 3 ? 1 : 0);
       

        

        switch ((state.board.piecesLeft[0] == 3 ? 1 : 0) + 
                (state.board.piecesLeft[1] == 3 ? 1 : 0))
        {

        case 0:
            int playerMobility = BoardAnalyser.calculateMobility(state.board, player, state.features);
            int opponentMobility = BoardAnalyser.calculateMobility(state.board, 1-player, state.features);
            int mobilityAdvantageScore = (playerMobility-opponentMobility) << BoardAnalyser.MOBILITY_ADVANTAGE_EXPONENT;

            if (playerMobility == 0 && board.unplacedPieces[player] == 0)
                return -BoardAnalyser.WIN_VALUE;
            else if (opponentMobility == 0 && board.unplacedPieces[otherPlayer] == 0) 
                return BoardAnalyser.WIN_VALUE;

            score += mobilityAdvantageScore;

            score += BoardAnalyser.calculateIntersectionAdvantageScore(board, player);
            score += BoardAnalyser.calculateRowAdvantage(board, player, freebb);

            break;

        case 1:
            score <<= BoardAnalyser.SINGLE_JUMPER_EXPONENT;

            score += BoardAnalyser.calculateIntersectionAdvantageScore(board, player);
            score += BoardAnalyser.calculateRowAdvantage(board, player, freebb);
        default:

        }

        return score;
    }

    public static int calculatePieceAdvantage(Board board, int player)
    {
        return board.piecesLeft[player] - board.piecesLeft[1-player];
    }

    public static int calculateIntersectionAdvantageScore(Board board, int player)
    {
        return Integer.bitCount(board.bbs[player] & THREE_WAY_INTERSECTIONS_BB) * THREE_WAY_INTERSECTION_VALUE 
             + Integer.bitCount(board.bbs[player] & FOUR_WAY_INTERSECTIONS_BB) * FOUR_WAY_INTERSECTION_VALUE;
    }

    public static int calculateRowAdvantage(Board board, int player, int freebb)
    {
        int score = 0;
        int mybb = board.bbs[player];
        int herbb = board.bbs[1 - player];
        int myrow, herrow;

        for (int millno = 0; millno <= 15; millno++)
        {
            int millmask = BoardFeatures.MILL_POSITION_BBS[millno];
            myrow = mybb & millmask;
            herrow = herbb & millmask;


            if (myrow > 0 && herrow == 0)
            {
                log.debug("Bonus - Player only one in row " + millno + ":");
                score += calculateRowValue(board, player, player, freebb, millno, millmask, mybb, myrow);
            }

            else if (herrow > 0 && myrow == 0)
            {
                log.debug("Penalty - Opponent only one in row " + millno + ":");
                score -= calculateRowValue(board, 1-player, player, freebb, millno, millmask, herbb, herrow);
            }
        }
        return score;
    }

    private static int calculateRowValue(Board board, int player, int currentPlayer, 
                                         int freebb, int millno, int millmask, int bb, int rowbb)
    {
        int rowfreebb = millmask & ~rowbb; 

        int score = 0;

        
        if (millno >= BoardFeatures.SPOKES_IDX)
        {
            log.debug("\tExclusivity on a horizontal or vertical.");
            score += SPOKE_EXCLUSIVITY_VALUE;
        }
        
        else if (millno >= BoardFeatures.MIDDLE_RING_IDX && millno < BoardFeatures.INNER_RING_IDX)
        {
            log.debug("\tExclusivity on a side of middle ring.");
            score += MIDDLE_RING_EXCLUSIVITY_VALUE;
        }

        switch (Integer.bitCount(rowbb))
        {
        case 3:
            score += MILL_VALUE;
            score += calculateMobility(freebb, rowbb) << MILL_MOBILITY_EXPONENT;
            log.debug("\tMill found.  Mobility: " + calculateMobility(freebb, rowbb));
            break;
        case 2:
                    if (board.unplacedPieces[player] > 0    
                || canSlideIntoRow(millmask, bb, rowfreebb, rowbb, player)) 
            {
                if (player == currentPlayer)
                {
                    log.debug("\tMill may be formed on THIS turn.");
                    score += POTENTIAL_MILL_VALUE;
                }
                else
                {
                    log.debug("\tMill may be formed on NEXT turn.");
                    score += POTENTIAL_MILL_NEXT_TURN_VALUE;
                }
            }
            else
                log.debug("\tMill cannot be formed by next turn.");
        case 1:
        
            score += ROW_EXCLUSIVITY_VALUE;
            break;
        }

        return score;
    }

    public static boolean canSlideIntoRow(int millmask, int bb, int rowfreebb, int rowbb, int player)
    {
        int notinrowbb = ~rowbb;
        
        for (int pos = 0; rowfreebb > 0; rowfreebb >>>= 1, pos++)
            if ((rowfreebb & 1) != 0)
            {
        
                int adjbb = MoveGenerator.ADJACENT_POSITION_BBS[pos] & notinrowbb;

        
                        if ((bb & adjbb) != 0) return true;

            }
        
        return false;
    }
    
    public static int calculateMobility(Board board, int player, BoardFeatures features)
    {
        return calculateMobility(features.freebb, board.bbs[player]);
    }
    
    private static int calculateMobility(int freebb, int bb)
    {
        int m = 0;
        int adjbb;

     
        for (int pos = 0; bb != 0; bb >>>= 1, pos++)
     
            if ((bb & 1) != 0)
            {
    
                m += Integer.bitCount(MoveGenerator.ADJACENT_POSITION_BBS[pos] & freebb);
            }

        return m;
    }

    public static void logBoardEvaluation(GameState state)
    {
        int player = state.currentPlayer;
        log.debug("Piece advantage:\t\t\t" + calculatePieceAdvantage(state.board, player));

        log.debug("Current player's mobility:\t\t" + calculateMobility(state.features.freebb, state.board.bbs[player]));
        log.debug("Opponent's mobility:\t\t" + calculateMobility(state.features.freebb, state.board.bbs[1 - player]));
        log.debug("Player's position value:\t\t" + calculateIntersectionAdvantageScore(state.board, player));
        log.debug("Opponent's position value:\t\t" + calculateIntersectionAdvantageScore(state.board, 1 - player));
        log.debug("Analysis score:\t\t"+BoardAnalyser.calculateScore(state));
    }
    
    public static void deactivateLog()
    {
        BoardAnalyser.log.setLevel(Level.OFF);
    }
    
    public static void reactivateLog()
    {
        BoardAnalyser.log.setLevel(originalLogLevel);
    }
}


